import 'dart:convert';

import 'package:code_climate/src/utils.dart';

/// A Code Climate issue represents a single instance of a real
/// or potential code problem, detected by a static analysis Engine.
/// Specification: https://github.com/codeclimate/platform/blob/master/spec/analyzers/SPEC.md#issues
class Issue {
  Issue({
    required this.checkName,
    required this.description,
    required this.location,
    required this.severity,
    required this.fingerprint,
    required this.categories,
    this.content,
  });

  /// Dart Analysis Server Specification tells us that dartanalyzer produces
  /// machine output in format `AnalysisErrorSeverity`|`AnalysisErrorType`|
  /// `code`|`path`|`line`|`column`|`length`|`description` where:
  ///
  /// * `AnalysisErrorSeverity` is an enumeration of the possible severities
  /// of analysis errors (INFO, WARNING, or ERROR)
  /// * `AnalysisErrorType` is an enumeration of the possible types
  /// of analysis errors (CHECKED_MODE_COMPILE_TIME_ERROR, COMPILE_TIME_ERROR,
  // ignore: todo
  /// HINT, LINT, STATIC_TYPE_WARNING, STATIC_WARNING, SYNTACTIC_ERROR, or TODO)
  /// * `code` is the name, as string, of the error code associated with
  /// this error.
  /// * `path` is the file containing the analysis error.
  /// * `line` is the one-based index of the line containing the first character
  /// of the range.
  /// * `column` is the offset of the range.
  /// * `length` is the length of the range.
  /// * `description` is the message to be displayed for this error. The message
  /// should indicate what is wrong with the code and why it is wrong.
  ///
  /// Dart Analysis Server API Specification: https://htmlpreview.github.io/?https://github.com/dart-lang/sdk/blob/master/pkg/analysis_server/doc/api.html
  factory Issue.fromAnalyzer(String analyzerLine) {
    final segments = analyzerLine.split('|');

    final line = int.parse(segments[4]);
    final beginColumn = int.parse(segments[5]);
    final endColumn = beginColumn + int.parse(segments[6]);

    return Issue(
      severity: parseSeverity(segments[0]),
      checkName: segments[2],
      description: segments[7],
      fingerprint: md5(analyzerLine),
      categories: [],
      location: Location(
        path: relativizePath(segments[3]),
        beginLine: line,
        beginColumn: beginColumn,
        endLine: endColumn,
        endColumn: line,
      ),
    );
  }

  Map<String, dynamic> toJson() => {
        'type': type,
        'check_name': checkName,
        'description': description,
        'location': location.toJson(),
        'severity': severity,
        'fingerprint': fingerprint,
        // if (content != null) 'content': content,
      };

  @override
  String toString() => jsonEncode(toJson());

  /// Must always be "issue".
  static const String type = 'issue';

  /// A unique name representing the static analysis check
  /// that emitted this issue.
  String checkName;

  /// A string explaining the issue that was detected.
  String description;

  /// A markdown snippet describing the issue,
  /// including deeper explanations and links to other resources.
  String? content;

  /// A Location object representing the place in the source code
  /// where the issue was discovered.
  Location location;

  /// A Severity string (info, minor, major, critical, or blocker)
  /// describing the potential impact of the issue found.
  String severity;

  /// A unique, deterministic identifier for the specific issue
  /// being reported to allow a user to exclude it from future analyses.
  String fingerprint;

  List<String> categories;
}

/// Location of the error:
/// See
/// https://github.com/codeclimate/platform/blob/master/spec/analyzers/SPEC.md#locations
/// Example JSON:
/// ```json
/// {
///   "path": "path/to/file.css",
///   "positions": {
///     "begin": {
///       "line": 3,
///       "column": 10
///     },
///     "end": {
///       "line": 4,
///       "column": 12
///     }
///   }
/// }
/// ```
class Location {
  final String path;
  final int beginLine;
  final int beginColumn;
  final int endLine;
  final int endColumn;

  Location({
    required this.path,
    required this.beginLine,
    required this.beginColumn,
    required this.endLine,
    required this.endColumn,
  });

  Map<String, dynamic> toJson() {
    return {
      'path': path,
      'positions': {
        'begin': {
          'line': beginLine,
          'column': beginColumn,
        },
        'end': {
          'line': endLine,
          'column': endColumn,
        },
      },
    };
  }
}
